/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.flink.table.plan.nodes

import org.apache.calcite.rel.core.JoinRelType
import org.apache.calcite.rex.{RexCall, RexInputRef, RexNode}
import org.apache.flink.api.common.functions.RichFlatMapFunction
import org.apache.flink.configuration.Configuration
import org.apache.flink.streaming.api.operators.OneInputStreamOperator
import org.apache.flink.table.functions.python.PythonFunctionInfo
import org.apache.flink.table.plan.nodes.CommonPythonCorrelate.{PYTHON_TABLE_FUNCTION_OPERATOR_NAME, PYTHON_TABLE_FUNCTION_FLAT_MAP_NAME}
import org.apache.flink.table.runtime.types.CRow
import org.apache.flink.table.types.logical.RowType
import org.apache.flink.types.Row

import scala.collection.mutable

trait CommonPythonCorrelate extends CommonPythonBase {
  protected def getPythonTableFunctionOperator(
      config: Configuration,
      inputRowType: RowType,
      outputRowType: RowType,
      pythonFunctionInfo: PythonFunctionInfo,
      udtfInputOffsets: Array[Int],
      joinType: JoinRelType): OneInputStreamOperator[CRow, CRow] = {
    val clazz = loadClass(PYTHON_TABLE_FUNCTION_OPERATOR_NAME)
    val ctor = clazz.getConstructor(
      classOf[Configuration],
      classOf[PythonFunctionInfo],
      classOf[RowType],
      classOf[RowType],
      classOf[Array[Int]],
      classOf[JoinRelType])
    ctor.newInstance(
      config,
      pythonFunctionInfo,
      inputRowType,
      outputRowType,
      udtfInputOffsets,
      joinType)
      .asInstanceOf[OneInputStreamOperator[CRow, CRow]]
  }

  protected def getPythonTableFunctionFlatMap(
      config: Configuration,
      inputRowType: RowType,
      outputRowType: RowType,
      pythonFunctionInfo: PythonFunctionInfo,
      udtfInputOffsets: Array[Int],
      joinType: JoinRelType): RichFlatMapFunction[Row, Row] = {
    val clazz = loadClass(PYTHON_TABLE_FUNCTION_FLAT_MAP_NAME)
    val ctor = clazz.getConstructor(
      classOf[Configuration],
      classOf[PythonFunctionInfo],
      classOf[RowType],
      classOf[RowType],
      classOf[Array[Int]],
      classOf[JoinRelType])
    ctor.newInstance(
      config,
      pythonFunctionInfo,
      inputRowType,
      outputRowType,
      udtfInputOffsets,
      joinType)
      .asInstanceOf[RichFlatMapFunction[Row, Row]]
  }

  protected def extractPythonTableFunctionInfo(
      pythonRexCall: RexCall): (Array[Int], PythonFunctionInfo) = {
    val inputNodes = new mutable.LinkedHashMap[RexNode, Integer]()
    val pythonTableFunctionInfo = createPythonFunctionInfo(pythonRexCall, inputNodes)
    val udtfInputOffsets = inputNodes.toArray
      .map(_._1)
      .collect { case inputRef: RexInputRef => inputRef.getIndex }
    (udtfInputOffsets, pythonTableFunctionInfo)
  }
}

object CommonPythonCorrelate {
  val PYTHON_TABLE_FUNCTION_OPERATOR_NAME =
    "org.apache.flink.table.runtime.operators.python.table.PythonTableFunctionOperator"
  val PYTHON_TABLE_FUNCTION_FLAT_MAP_NAME =
    "org.apache.flink.table.runtime.functions.python.PythonTableFunctionFlatMap"
}
